import http.client
import json
import time
import os

from .web_service import WebService
from .electronic_billing import ElectronicBilling
from .register_inscription_proof import RegisterInscriptionProof
from .register_scope_ten import RegisterScopeTen
from .register_scope_thirteen import RegisterScopeThirteen

os.environ["SSL_CERT_FILE"] = os.path.join(os.path.dirname(__file__), "data", "cacert.pem")

class Afip:
  sdk_version_number = '1.1.1'

  def __init__(self, options: dict):
    if not(options.get("CUIT")):
      raise Exception("CUIT field is required in options")

    self.CUIT: int = options.get("CUIT")
    self.production: bool = options.get("production") if options.get("production") == True else False
    self.environment: str = "prod" if self.production == True else "dev"
    self.cert: str = options.get("cert")
    self.key: str = options.get("key")
    self.access_token: str = options.get("access_token")
  
    self.ElectronicBilling = ElectronicBilling(self)
    self.RegisterInscriptionProof = RegisterInscriptionProof(self)
    self.RegisterScopeTen = RegisterScopeTen(self)
    self.RegisterScopeThirteen = RegisterScopeThirteen(self)


  # Gets token authorization for an AFIP Web Service
  #
  # If force is true it forces to create a new TA
  def getServiceTA(self, service: str, force: bool = False) -> dict:
    conn = http.client.HTTPSConnection("app.afipsdk.com")

    payload = {
      "environment": self.environment,
      "tax_id": self.CUIT,
      "wsid": service,
      "force_create": force
    }

    if self.cert: payload["cert"] = self.cert
    if self.key: payload["key"] = self.key

    headers = {
      "Content-Type": "application/json",
      "sdk-version-number": self.sdk_version_number,
      "sdk-library": "python",
      "sdk-environment": self.environment
    }

    if self.access_token: headers["Authorization"] = "Bearer %s" % self.access_token

    conn.request("POST", "/api/v1/afip/auth", json.dumps(payload), headers)

    res = conn.getresponse()
    
    data = res.read()

    if res.getcode() >= 400:
      raise Exception(data.decode("utf-8"))

    return json.loads(data.decode("utf-8"))
  
  # Get last request and last response XML
  def getLastRequestXML(self) -> dict:
    conn = http.client.HTTPSConnection("app.afipsdk.com")

    headers = {
      "Content-Type": "application/json",
      "sdk-version-number": self.sdk_version_number,
      "sdk-library": "python",
      "sdk-environment": self.environment
    }

    if self.access_token: headers["Authorization"] = "Bearer %s" % self.access_token

    conn.request("GET", "/api/v1/afip/requests/last-xml", None, headers)

    res = conn.getresponse()
    
    data = res.read()

    if res.getcode() >= 400:
      raise Exception(data.decode("utf-8"))

    return json.loads(data.decode("utf-8"))
  

  # Create generic Web Service
  def webService(self, service, options: dict = {}) -> WebService:
    options['service'] = service
    options['generic'] = True

    return WebService(self, options)
  
  # Create AFIP cert
  def createCert(self, username: str, password: str, alias: str) -> dict:
    conn = http.client.HTTPSConnection("app.afipsdk.com")

    payload = {
      "environment": self.environment,
      "tax_id": self.CUIT,
      "username": username,
      "password": password,
      "alias": alias
    }

    headers = {
      "Content-Type": "application/json",
      "sdk-version-number": self.sdk_version_number,
      "sdk-library": "python",
      "sdk-environment": self.environment
    }

    if self.access_token: headers["Authorization"] = "Bearer %s" % self.access_token

    retry_request = 24

    while retry_request >= 0:
      retry_request -= 1

      conn.request("POST", "/api/v1/afip/certs", json.dumps(payload), headers)

      res = conn.getresponse()
      
      data = res.read()

      if res.getcode() >= 400:
        raise Exception(data.decode("utf-8"))
    
      decoded_res = json.loads(data.decode("utf-8"))

      if decoded_res["status"] == "complete":
        return decoded_res["data"]

      if decoded_res.get("long_job_id"):
        payload["long_job_id"] = decoded_res["long_job_id"]

      time.sleep(5)

    raise Exception("Error: Waiting for too long")
  
  # Create AFIP cert
  def createWSAuth(self, username: str, password: str, alias: str, wsid: str) -> dict:
    conn = http.client.HTTPSConnection("app.afipsdk.com")

    payload = {
      "environment": self.environment,
      "tax_id": self.CUIT,
      "username": username,
      "password": password,
      "alias": alias,
      "wsid": wsid
    }

    headers = {
      "Content-Type": "application/json",
      "sdk-version-number": self.sdk_version_number,
      "sdk-library": "python",
      "sdk-environment": self.environment
    }

    if self.access_token: headers["Authorization"] = "Bearer %s" % self.access_token

    retry_request = 24

    while retry_request >= 0:
      retry_request -= 1

      conn.request("POST", "/api/v1/afip/ws-auths", json.dumps(payload), headers)

      res = conn.getresponse()
      
      data = res.read()

      if res.getcode() >= 400:
        raise Exception(data.decode("utf-8"))
    
      decoded_res = json.loads(data.decode("utf-8"))

      if decoded_res["status"] == "complete":
        return decoded_res["data"]
      
      if decoded_res.get("long_job_id"):
        payload["long_job_id"] = decoded_res["long_job_id"]

      time.sleep(5)
      

    raise Exception("Error: Waiting for too long")
  


import http.client
import json

class WebService:
  def __init__(self, afip, options: dict = {}):
    self.afip = afip
    self.options = options
    self.WSDL = options.get('WSDL')
    self.URL = options.get('URL')
    self.WSDL_TEST = options.get('WSDL_TEST')
    self.URL_TEST = options.get('URL_TEST')
    self.soapv12 = options.get('soapV1_2')

    if not(options.get('service')):
      raise Exception("service field is required in options")
    
  # Gets token authorization for an AFIP Web Service
  #
  # If force is true it forces to create a new TA
  def getTokenAuthorization(self, force = False) -> dict:
    return self.afip.getServiceTA(self.options.get('service'), force)

  # Sends request to AFIP servers
  def executeRequest(self, method, params = {}) -> dict:
    conn = http.client.HTTPSConnection("app.afipsdk.com")

    payload = {
      "method": method,
      "params": params,
      "environment": self.afip.environment,
      "wsid": self.options.get('service'),
      "url": self.URL if self.afip.production else self.URL_TEST,
      "wsdl": self.WSDL if self.afip.production else self.WSDL_TEST,
      "soap_v_1_2": self.soapv12
    }

    headers = {
      "Content-Type": "application/json",
      "sdk-version-number": self.afip.sdk_version_number,
      "sdk-library": "python",
      "sdk-environment": self.afip.environment
    }

    if self.afip.access_token: headers["Authorization"] = "Bearer %s" % self.afip.access_token

    conn.request("POST", "/api/v1/afip/requests", json.dumps(payload), headers)

    res = conn.getresponse()
    
    data = res.read()

    if res.getcode() >= 400:
      raise Exception(data.decode("utf-8"))

    return json.loads(data.decode("utf-8"))


import http.client
import json
import re

from .web_service import WebService

class ElectronicBilling(WebService):
  soapv12 = True
  WSDL = "https://servicios1.afip.gov.ar/wsfev1/service.asmx?WSDL"
  URL = "https://servicios1.afip.gov.ar/wsfev1/service.asmx"
  WSDL_TEST = "https://wswhomo.afip.gov.ar/wsfev1/service.asmx?WSDL"
  URL_TEST = "https://wswhomo.afip.gov.ar/wsfev1/service.asmx"

  def __init__(self, afip):
    super(ElectronicBilling, self).__init__(afip, { "service": "wsfe" })

  # Create PDF
  def createPDF(self, data: dict):
    conn = http.client.HTTPSConnection("app.afipsdk.com")

    headers = {
        "Content-Type": "application/json",
        "sdk-version-number": self.afip.sdk_version_number,
        "sdk-library": "python",
        "sdk-environment": self.afip.environment
    }

    if self.afip.access_token: headers["Authorization"] = "Bearer %s" % self.afip.access_token

    conn.request("POST", "/api/v1/pdfs", json.dumps(data), headers)

    res = conn.getresponse()
    
    data = res.read()

    if res.getcode() >= 400:
        raise Exception(data.decode("utf-8"))

    response_data = json.loads(data.decode("utf-8"))

    return { 
      "file": response_data["file"], 
      "file_name": response_data["file_name"]
    }

  # Gets last voucher number
  def getLastVoucher(self, sales_point: int, type: int):
    req = {
      "PtoVta": sales_point,
      "CbteTipo": type
    }

    return self.executeRequest("FECompUltimoAutorizado", req)["CbteNro"]

  # Create a voucher from AFIP
  def createVoucher(self, data: dict, return_response: bool = False):
    # Reassign data to avoid modify te original object
    data = data.copy()

    req = {
      "FeCAEReq": {
        "FeCabReq": {
          "CantReg": data["CbteHasta"] - data["CbteDesde"] + 1,
          "PtoVta": data["PtoVta"],
          "CbteTipo": data["CbteTipo"]
        },
        "FeDetReq": {
          "FECAEDetRequest": data
        }
      }
    }

    data.pop("CantReg")
    data.pop("PtoVta")
    data.pop("CbteTipo")

    if data.get("Tributos"): data["Tributos"] = { "Tributo": data["Tributos"] }
    if data.get("Iva"): data["Iva"] = { "AlicIva": data["Iva"] }
    if data.get("CbtesAsoc"): data["CbtesAsoc"] = { "CbteAsoc": data["CbtesAsoc"] }
    if data.get("Compradores"): data["Compradores"] = { "Comprador": data["Compradores"] }
    if data.get("Opcionales"): data["Opcionales"] = { "Opcional": data["Opcionales"] }

    results = self.executeRequest("FECAESolicitar", req)

    if return_response == True:
      return results
    
    if type(results["FeDetResp"]["FECAEDetResponse"]) in (tuple, list):
      results["FeDetResp"]["FECAEDetResponse"] = results["FeDetResp"]["FECAEDetResponse"][0]

    return {
      "CAE": results["FeDetResp"]["FECAEDetResponse"]["CAE"],
      "CAEFchVto": self.formatDate(results["FeDetResp"]["FECAEDetResponse"]["CAEFchVto"])
    }

  # Create next voucher from AFIP
  def createNextVoucher(self, data: dict):
    # Reassign data to avoid modify te original object
    data = data.copy()

    lastVoucher = self.getLastVoucher(data["PtoVta"], data["CbteTipo"])

    voucherNumber = lastVoucher + 1

    data["CbteDesde"] = voucherNumber
    data["CbteHasta"] = voucherNumber

    res = self.createVoucher(data)

    res["voucherNumber"] = voucherNumber

    return res

  # Get complete voucher information
  def getVoucherInfo(self, number: int, sales_point: int, type: int):
    req = {
      "FeCompConsReq": {
        "CbteNro": number,
        "PtoVta": sales_point,
        "CbteTipo": type
      }
    }

    return self.executeRequest("FECompConsultar", req)

  # Create CAEA
  def createCAEA(self, period: int, fortnight: int):
    req = {
      "Periodo": period,
      "Orden": fortnight
    }

    return self.executeRequest("FECAEASolicitar", req)["ResultGet"]

  # Get CAEA
  def getCAEA(self, period: int, fortnight: int):
    req = {
      "Periodo": period,
      "Orden": fortnight
    }

    return self.executeRequest("FECAEAConsultar", req)["ResultGet"]

  # Asks to AFIP Servers for available sales points
  def getSalesPoints(self):
    return self.executeRequest("FEParamGetPtosVenta")["ResultGet"]["PtoVenta"]

  # Asks to AFIP Servers for available voucher types
  def getVoucherTypes(self):
    return self.executeRequest("FEParamGetTiposCbte")["ResultGet"]["CbteTipo"]

  # Asks to AFIP Servers for voucher concepts availables
  def getConceptTypes(self):
    return self.executeRequest("FEParamGetTiposConcepto")["ResultGet"]["ConceptoTipo"]

  # Asks to AFIP Servers for document types availables
  def getDocumentTypes(self):
    return self.executeRequest("FEParamGetTiposDoc")["ResultGet"]["DocTipo"]

  # Asks to AFIP Servers for available aliquotes
  def getAliquotTypes(self):
    return self.executeRequest("FEParamGetTiposIva")["ResultGet"]["IvaTipo"]

  # Asks to AFIP Servers for available currencies
  def getCurrenciesTypes(self):
    return self.executeRequest("FEParamGetTiposMonedas")["ResultGet"]["Moneda"]

  # Asks to AFIP Servers for available voucher optional data
  def getOptionsTypes(self):
    return self.executeRequest("FEParamGetTiposOpcional")["ResultGet"]["OpcionalTipo"]

  # Asks to AFIP Servers for available tax types
  def getTaxTypes(self):
    return self.executeRequest("FEParamGetTiposTributos")["ResultGet"]["TributoTipo"]

  # Asks to web service for servers status
  def getServerStatus(self):
    return self.executeRequest("FEDummy")

  # Change date from AFIP used format (yyyymmdd) to yyyy-mm-dd
  def formatDate(self, date: int) -> str:
    m = re.search(r"(\d{4})(\d{2})(\d{2})", str(date))

    return "%s-%s-%s" %(m.group(1), m.group(2), m.group(3))

  # Sends request to AFIP servers
  def executeRequest(self, operation: str, params: dict = {}):
    params.update(self.getWSInitialRequest(operation))

    results = super(ElectronicBilling, self).executeRequest(operation, params)

    self.__checkErrors(operation, results)

    return results["%sResult" % operation]

  # Prepare default request parameters for most operations
  def getWSInitialRequest(self, operation: str):
    if operation == "FEDummy":
      return {}
    
    ta = self.getTokenAuthorization()

    return {
      "Auth": {
        "Token": ta["token"],
        "Sign": ta["sign"],
        "Cuit": self.afip.CUIT
      }
    }

  # Check if occurs an error on Web Service request
  def __checkErrors(self, operation: str, results: dict):
    res = results["%sResult" %operation]

    if operation == "FECAESolicitar" and res["FeDetResp"]:
      if type(res["FeDetResp"]["FECAEDetResponse"]) in (tuple, list):
        res["FeDetResp"]["FECAEDetResponse"] = res["FeDetResp"]["FECAEDetResponse"][0]

      if res["FeDetResp"]["FECAEDetResponse"].get("Observaciones") and res["FeDetResp"]["FECAEDetResponse"]["Resultado"] != "A":
        res["Errors"] = { "Err": res["FeDetResp"]["FECAEDetResponse"]["Observaciones"]["Obs"] }

    if res.get("Errors"):
      err = res["Errors"]["Err"][0] if type(res["Errors"]["Err"]) in (tuple, list) else res["Errors"]["Err"]

      raise Exception("(%s) %s" % (err["Code"], err["Msg"])) 



from .web_service import WebService

class RegisterInscriptionProof(WebService):
  soapv12 = False
  WSDL = "https://aws.afip.gov.ar/sr-padron/webservices/personaServiceA5?WSDL"
  URL = "https://aws.afip.gov.ar/sr-padron/webservices/personaServiceA5"
  WSDL_TEST = "https://awshomo.afip.gov.ar/sr-padron/webservices/personaServiceA5?WSDL"
  URL_TEST = "https://awshomo.afip.gov.ar/sr-padron/webservices/personaServiceA5"
  
  def __init__(self, afip):
    super(RegisterInscriptionProof, self).__init__(afip, { "service": "ws_sr_constancia_inscripcion" })

  # Asks to web service for taxpayer details
  def getTaxpayerDetails(self, identifier: int):
    # Get token and sign
    ta = self.getTokenAuthorization()

    # Prepare params
    params = {
      "token": ta["token"],
      "sign": ta["sign"],
      "cuitRepresentada": self.afip.CUIT,
      "idPersona": identifier
    }

    return self.executeRequest("getPersona_v2", params)

  # Asks to web service for taxpayers details
  def getTaxpayersDetails(self, identifiers: list):
    # Get token and sign
    ta = self.getTokenAuthorization()

    # Prepare params
    params = {
      "token": ta["token"],
      "sign": ta["sign"],
      "cuitRepresentada": self.afip.CUIT,
      "idPersona": identifiers
    }

    return self.executeRequest("getPersonaList_v2", params)["persona"]

  # Asks to web service for servers status
  def getServerStatus(self):
    return self.executeRequest("dummy")

  # Send request to AFIP servers
  def executeRequest(self, operation, params = {}):
    results = super(RegisterInscriptionProof, self).executeRequest(operation, params)

    if operation == "getPersona_v2":
      return results["personaReturn"]
    elif operation == "getPersonaList_v2":
      return results["personaListReturn"]
    else:
      return results["return"]



from .web_service import WebService

class RegisterScopeTen(WebService):
  soapv12 = False
  WSDL = "https://aws.afip.gov.ar/sr-padron/webservices/personaServiceA10?WSDL"
  URL = "https://aws.afip.gov.ar/sr-padron/webservices/personaServiceA10"
  WSDL_TEST = "https://awshomo.afip.gov.ar/sr-padron/webservices/personaServiceA10?WSDL"
  URL_TEST = "https://awshomo.afip.gov.ar/sr-padron/webservices/personaServiceA10"

  def __init__(self, afip):
    super(RegisterScopeTen, self).__init__(afip, { "service": "ws_sr_padron_a10" })

  # Asks to web service for taxpayer details
  def getTaxpayerDetails(self, identifier: int):
    # Get token and sign
    ta = self.getTokenAuthorization()

    # Prepare params
    params = {
      "token": ta["token"],
      "sign": ta["sign"],
      "cuitRepresentada": self.afip.CUIT,
      "idPersona": identifier
    }

    return self.executeRequest("getPersona", params)

  # Asks to web service for servers status
  def getServerStatus(self):
    return self.executeRequest("dummy")

  # Send request to AFIP servers
  def executeRequest(self, operation, params = {}):
    results = super(RegisterScopeTen, self).executeRequest(operation, params)

    if operation == "getPersona":
      return results["personaReturn"]
    else:
      return results["return"]
